/*
 * "Small Hello World" example.
 *
 * This example prints 'Hello from Nios II' to the STDOUT stream. It runs on
 * the Nios II 'standard', 'full_featured', 'fast', and 'low_cost' example
 * designs. It requires a STDOUT  device in your system's hardware.
 *
 * The purpose of this example is to demonstrate the smallest possible Hello
 * World application, using the Nios II HAL library.  The memory footprint
 * of this hosted application is ~332 bytes by default using the standard
 * reference design.  For a more fully featured Hello World application
 * example, see the example titled "Hello World".
 *
 * The memory footprint of this example has been reduced by making the
 * following changes to the normal "Hello World" example.
 * Check in the Nios II Software Developers Manual for a more complete
 * description.
 *
 * In the SW Application project (small_hello_world):
 *
 *  - In the C/C++ Build page
 *
 *    - Set the Optimization Level to -Os
 *
 * In System Library project (small_hello_world_syslib):
 *  - In the C/C++ Build page
 *
 *    - Set the Optimization Level to -Os
 *
 *    - Define the preprocessor option ALT_NO_INSTRUCTION_EMULATION
 *      This removes software exception handling, which means that you cannot
 *      run code compiled for Nios II cpu with a hardware multiplier on a core
 *      without a the multiply unit. Check the Nios II Software Developers
 *      Manual for more details.
 *
 *  - In the System Library page:
 *    - Set Periodic system timer and Timestamp timer to none
 *      This prevents the automatic inclusion of the timer driver.
 *
 *    - Set Max file descriptors to 4
 *      This reduces the size of the file handle pool.
 *
 *    - Check Main function does not exit
 *    - Uncheck Clean exit (flush buffers)
 *      This removes the unneeded call to exit when main returns, since it
 *      won't.
 *
 *    - Check Don't use C++
 *      This builds without the C++ support code.
 *
 *    - Check Small C library
 *      This uses a reduced functionality C library, which lacks
 *      support for buffering, file IO, floating point and getch(), etc.
 *      Check the Nios II Software Developers Manual for a complete list.
 *
 *    - Check Reduced device drivers
 *      This uses reduced functionality drivers if they're available. For the
 *      standard design this means you get polled UART and JTAG UART drivers,
 *      no support for the LCD driver and you lose the ability to program
 *      CFI compliant flash devices.
 *
 *    - Check Access device drivers directly
 *      This bypasses the device file system to access device drivers directly.
 *      This eliminates the space required for the device file system services.
 *      It also provides a HAL version of libc services that access the drivers
 *      directly, further reducing space. Only a limited number of libc
 *      functions are available in this configuration.
 *
 *    - Use ALT versions of stdio routines:
 *
 *           Function                  Description
 *        ===============  =====================================
 *        alt_printf       Only supports %s, %x, and %c ( < 1 Kbyte)
 *        alt_putstr       Smaller overhead than puts with direct drivers
 *                         Note this function doesn't add a newline.
 *        alt_putchar      Smaller overhead than putchar with direct drivers
 *        alt_getchar      Smaller overhead than getchar with direct drivers
 *
 */

#include "system.h"
#include <unistd.h>
#include <stdio.h>
#include "altera_avalon_pio_regs.h"
#include "altera_avalon_lcd_16207_regs.h"

void lcd_init( void );
void test_lcd( void );


void lcd_init( void ) {
	/* Set Function Code Four Times -- 8-bit, 2 line, 5x7 mode */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x38);
	//IOWR( LCD_BASE, LCD_WR_COMMAND_REG, 0x38 );
	usleep(4100); //* Wait 4.1 ms */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x38 );
	usleep(100); /* Wait 100 us */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x38 );
	usleep(5000); /* Wait 5.0 ms */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x38 );
	usleep(100);

	/* Set Display to OFF */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x08 );
	usleep(100);

	/* Set Display to ON */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x0C );
	usleep(100);

	/* Set Entry Mode -- Cursor increment, display doesn't shift */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x06 );
	usleep(100);

	/* Set the cursor to the home position */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x02 );
	usleep(2000);

	/* Clear the display */
	IOWR_ALTERA_AVALON_LCD_16207_COMMAND(LCD_0_BASE, 0x01 );
	usleep(2000);
}


void lcd_string_write(char string[],char line)
{
   int i = 0;

   if(line == 0) //If 'line' is 0 it means line 1
   {
       /* Set the cursor to the home position */
       IOWR_ALTERA_AVALON_LCD_16207_COMMAND( LCD_0_BASE, 0x02 );
       usleep(2000);
       for(i=0;i<16;i++)
       {
           if(string[i] == '\0')
           {
               for(;i<16;i++)
               {
                   IOWR_ALTERA_AVALON_LCD_16207_DATA(LCD_0_BASE,' ');
                   usleep(100);
               }
               break;
           }
           IOWR_ALTERA_AVALON_LCD_16207_DATA(LCD_0_BASE,string[i]);
           usleep(100);
       }
   }
   else //If 'line' is other than 0 use line 2
   {
       /* Count along the bottom row */
       /* Set Address */
       IOWR_ALTERA_AVALON_LCD_16207_COMMAND( LCD_0_BASE, 0xC0 );
       usleep(1000);

       for(i=0;i<16;i++)
       {
           if(string[i] == '\0')
           {
               for(;i<16;i++)
               {
                   IOWR_ALTERA_AVALON_LCD_16207_DATA(LCD_0_BASE,' ');
                   usleep(100);
               }
               break;
           }
           IOWR_ALTERA_AVALON_LCD_16207_DATA(LCD_0_BASE,string[i]);
           usleep(100);
       }

   }

}


int main(void){
	lcd_string_write("Embedded audio",0); //Writes Hello, on line 1
	lcd_string_write("demonstrator",1); //Writes Bye, on line 2
	usleep(3000000); /* Wait 3.0 s */
	while(1){
		if((IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) & 0x01) == 0x00){
			lcd_string_write("Sample rate",0);
			lcd_string_write("is updated...",1);
			usleep(1000000); /* Wait 1.0 s */
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x01){
			lcd_string_write("Sample: 8kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x03){
			lcd_string_write("Sample: 32kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x05){
			lcd_string_write("Sample: 48kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x07){
			lcd_string_write("Sample: 96kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x09){
			lcd_string_write("Sample: 8kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x0B){
			lcd_string_write("Sample: 32kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x0D){
			lcd_string_write("Sample: 48kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x0F){
			lcd_string_write("Sample: 96kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Off",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x11){
			lcd_string_write("Sample: 8kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Lowpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x13){
			lcd_string_write("Sample: 32kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Lowpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x15){
			lcd_string_write("Sample: 48kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Lowpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x17){
			lcd_string_write("Sample: 96kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Lowpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x19){
			lcd_string_write("Sample: 8kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Highpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x1B){
			lcd_string_write("Sample: 32kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Highpass",1); //Writes Bye, on line 2
		}

		else if(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x1D){
			lcd_string_write("Sample: 48kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Highpass",1); //Writes Bye, on line 2
		}

		else{//(IORD_ALTERA_AVALON_PIO_DATA(PIO_0_BASE) == 0x1F){
			lcd_string_write("Sample: 96kHz",0); //Writes Hello, on line 1
			lcd_string_write("Filter: Highpass",1); //Writes Bye, on line 2
		}

	}
}


